;
;  S3 Vertical Retrace Interrupt Manager
;
;  By Gmlb256
;

.8086
.model   small

cseg segment para public 'code'
     assume cs:cseg, ds:cseg, es:cseg

    org    100h
main:
    jmp    start

;
;  Constant values
;

CmdLength         equ   80h
CmdData           equ   81h
TsrIn             equ   'GM'
TsrOut            equ   'LB'

;
;  Resident portion of the utility
;

resident:

OldInt10_Service  label dword
OldInt10Offset    dw    ?
OldInt10Segment   dw    ?
Int10Offset       dw    0
Int10Segment      dw    0
InterruptEnabled  db    1

;
;  Macro to print a message into the screen
;
print macro string

    mov      dx, offset string
    mov      ah, 9
    int     21h

endm print

;
;  Macro to enable the vertical retrace interrupt
;
EnableInterrupt macro

    push     ax
    push     bx
    push     dx

    call    UnlockRegisters

    ; get CR32 value into the AL register
    mov      al, 32h
    out      dx, al
    inc      dx
    in       al, dx

    ; enable the vertical retrace interrupt
    or       al, 10h
    out      dx, al
    dec      dx

    call    LockRegisters

    pop      dx
    pop      bx
    pop      ax

endm EnableInterrupt

;
;  Macro to disable the vertical retrace interrupt
;
DisableInterrupt macro

    push     ax
    push     bx
    push     dx

    call    UnlockRegisters

    ; get CR32 value into the AL register
    mov      al, 32h
    out      dx, al
    inc      dx
    in       al, dx

    ; disable the vertical retrace interrupt
    and      al, NOT 10h
    out      dx, al
    dec      dx

    call    LockRegisters

    pop      dx
    pop      bx
    pop      ax

endm DisableInterrupt

;
;  UnlockRegisters
;
;  Save the CR38 value and unlock the extended CRTC registers CR2D-CR3F.
;
;  AL = temp value
;  BL = previous CR38 value
;  DX = VGA address port 0x3d4 and 0x3d5
;
public  UnlockRegisters
    UnlockRegisters proc

    mov      dx, 3d4h
    mov      al, 38h
    out      dx, al
    inc      dx
    in       al, dx
    mov      bl, al
    mov      al, 48h
    out      dx, al
    dec      dx
    ret

UnlockRegisters endp

;
;  LockRegisters
;
;  Lock the extended CRTC registers CR2D-CR3F with the previous CR38 value.
;
;  AL = temp value
;  BL = previous CR38 value
;  DX = VGA address port 0x3d4 and 0x3d5
;
public  LockRegisters
    LockRegisters proc

    mov      al, 38h
    out      dx, al
    mov      al, bl
    inc      dx
    out      dx, al
    ret

LockRegisters endp

;
;  Int10_Service
;
;  INT 10h replacement used to reactivate the vertial retrace interrupt after
;  changing the video mode.
;
public  Int10_Service
    Int10_Service proc

    ;
    ; AX=0xC600: TSR installation check
    ;
    ; returns
    ; AX = 0x0000
    ; DX = TsrOut
    ; ES = TSR code segment
    ;
    cmp      ax, 0c600h
    je      int10_3

    ; check if the video mode is being changed
    test     ah, ah
    jz      int10_1

    ;
    ; VESA video modes doesn't use the EGA/VGA vertical retrace interrupt
    ; but at least on S3 video cards it's used when operating in enhanced
    ; mode instead of VGA mode for things such as S3D engine, DMA and FIFO
    ; operations in addition to the vertical retrace (this is just in case)
    ;
    cmp      ax, 4f02h
    jne     int10_4

int10_1:
    ; first change the video mode
    pushf
    call    cs:OldInt10_Service

    ; is the vertical retrace interrupt enabled?
    cmp     cs:InterruptEnabled, 1
    jne     int10_2

    ; reenable the vertical retrace interrupt since it gets disabled after
    ; chaning the video mode
    EnableInterrupt
int10_2:
    iret

int10_3:
    cmp      dx, TsrIn
    jne     int10_4
    mov      dx, TsrOut
    mov      ax, cs
    mov      es, ax
    iret

int10_4:
    jmp     cs:OldInt10_Service

Int10_Service endp

;
;  Transient portion of the utility
;

start:
    print   intro_str

    ; this utility requires DOS 3.30 or greater
    mov      ax, 3000h
    int     21h
    cmp      al, 3
    jb      unsupported_dos
    ja      supported_dos

    ; check the minor version of DOS 3
    cmp      ah, 30
    jb      unsupported_dos

supported_dos:
    ; check TSR installation status by calling a custom INT 10H routine
    mov      ax, 0c600h
    mov      dx, TsrIn
    int     10h
    cmp      dx, TsrOut
    je      active
    xor      dx, dx

    ; check the command line parameter first
    call    CheckParam

    ;
    ; check if there is a S3 video card installed
    ;
    ; NOTE: S3 chips prior to the Trio32/64 doesn't need this utility since
    ; they don't have a interrupt switch at CR32, so they will be declared
    ; unsupported
    ;
    call    UnlockRegisters

    mov      dx, 3d4h
    mov      al, 30h
    out      dx, al
    inc      dx
    in       al, dx
    dec      dx
    mov      cl, al

    call    LockRegisters

    ; the chip ID must be 0xE1 in order to proceed
    cmp      cl, 0e1h
    jne     unsupported_vid

    ; get the current INT 10H routine
    push     es
    mov      ax, 3510h
    int     21h
    mov     OldInt10Segment, es
    mov     OldInt10Offset, bx
    pop      es

    cmp     InterruptEnabled, 0
    je      no_interrupt

    EnableInterrupt

no_interrupt:
    call    PrintStatus
    print   installed_str

    ; set the new INT 10H routine
    mov      ax, 2510h
    lea      dx, Int10_Service
    int     21h

    ; set the new INT 10H segment and offset for future use
    mov     Int10Segment, ds
    mov     Int10Offset, dx

    ; deallocate environment block
    push     es
    mov      es, ds:[2ch]
    mov      ah, 49h
    int     21h
    pop      es

    ; compute the size of the resident portion in paragraphs
    lea      dx, start
    add      dx, 15
    mov      cl, 4
    shr      dx, cl

    ; start the TSR
    mov      ax, 3100h
    int     21h

unsupported_dos:
    print   unsuppdos_str

    ; DOS 1.0 must use INT 20H to exit the program
    int     20h

unsupported_vid:
    print   unsuppvid_str
    jmp     exit_1

active:
    ; if the TSR is currently installed, only check the command line parameters
    ; and print the current status of the vertical retrace interrupt
    call    CheckParam
    call    PrintStatus

exit_1:
    mov      ah, 4ch
    int     21h

;
;  CheckParam
;
;  Only check the first command line parameter which is only one character
;  after the slash or a minus sign to keep it simple as possible.
;
public  CheckParam
    CheckParam proc near

    mov      al, byte ptr cs:[CmdLength]
    mov      bx, CmdData - 1
    test     al, al
    jz      done

is_space:
    inc      bx
    mov      al, byte ptr cs:[bx]
    cmp      al, ' '
    je      is_space
    cmp      al, 0dh
    je      done
    cmp      al, '/'
    je      valid_param
    cmp      al, '-'
    je      valid_param

    jmp     invalid_param

valid_param:
    ; the parameter must have only one character, if the parameter have any
    ; additional character other than the carriage return or the space it will
    ; be declared an invalid parameter
    mov      al, byte ptr cs:[bx + 2]
    cmp      al, ' '
    je      valid_param_2
    cmp      al, 0dh
    je      valid_param_2

    jmp     invalid_param

valid_param_2:
    ; if the character is an uppercase letter, convert it to lowercase
    mov      al, byte ptr cs:[bx + 1]
    cmp      al, 'A'
    jb      not_uppercase
    cmp      al, 'Z'
    ja      not_uppercase
    sub      al, 'A' - 'a'

not_uppercase:
    cmp      al, '?'
    je      show_help
    cmp      al, 'e'
    je      enable_interrupt
    cmp      al, 'd'
    je      disable_interrupt
    cmp      al, 'u'
    je      uninstall_tsr

invalid_param:
    print   invalid_str
    jmp     exit_2

show_help:
    print   help_str

exit_2:
    mov      ah, 4ch
    int     21h

uninstall_tsr:
    call    UninstallTSR

enable_interrupt:
    mov     es:InterruptEnabled, 1
    EnableInterrupt
    jmp     done

disable_interrupt:
    mov     es:InterruptEnabled, 0
    DisableInterrupt
done:
    ret

CheckParam endp

;
;  PrintStatus
;
;  Check if the vertical retrace interrupt is enabled or not and print the
;  status into the screen.
;
public  PrintStatus
    PrintStatus proc near

    print   status_str

    cmp     es:InterruptEnabled, 1
    jne     interrupt_disabled
    print   enabled_str
    ret

interrupt_disabled:
    print   disabled_str
    ret

PrintStatus endp

;
;  UninstallTSR
;
public  UninstallTSR
    UninstallTSR proc near

    ; check if the TSR is installed first
    test     dx, dx
    jz      not_installed

    ; get the current INT 10H routine
    push     es
    mov      ax, 3510h
    int     21h
    mov     Int10Segment, es
    mov     Int10Offset, bx
    pop      es

    ; check if the INT 10H routine in last chain still point to our routine
    mov      dx, Int10Segment
    cmp      dx, es:Int10Segment
    jne     not_last
    cmp      bx, es:Int10Offset
    jne     not_last

    ; disable the vertical retrace interrupt before uninstalling the TSR
    DisableInterrupt

    ; restore previous INT 10H routine
    push     ds
    mov      ax, 2510h
    lds      dx, es:OldInt10_Service
    int     21h
    pop      ds

    ; deallocate environment block from the TSR segment
    push     es
    mov      es, es:[2ch]
    mov      ah, 49h
    int     21h
    pop      es

    ; remove the TSR from memory
    mov      ah, 49h
    int     21h

    print   uninstalled_str
    jmp     exit_3

not_installed:
    print   notinstalled_str
    jmp     exit_3

not_last:
    print   notlast_str

exit_3:
    mov      ah, 4ch
    int     21h

UninstallTSR endp

intro_str        db  'S3 Vertical Retrace Interrupt Manager by Gmlb256',0dh,0ah,0dh,0ah,'$'
unsuppdos_str    db  'Unsupported DOS version',0dh,0ah,'$'
invalid_str      db  'Invalid parameter',0dh,0ah,'$'
help_str         db  '/E - Enable the vertical retrace interrupt (default)',0dh,0ah
                 db  '/D - Disable the vertical retrace interrupt',0dh,0ah
                 db  '/U - Uninstall',0dh,0ah,'$'
unsuppvid_str    db  'Unsupported video card',0dh,0ah,'$'
status_str       db  'Status: $'
enabled_str      db  'Enabled',0dh,0ah,'$'
disabled_str     db  'Disabled',0dh,0ah,'$'
installed_str    db  0dh,0ah,'Installed',0dh,0ah,'$'
notinstalled_str db  'Not installed',0dh,0ah,'$'
notlast_str      db  'Unable to uninstall, not last in chain!',0dh,0ah,'$'
uninstalled_str  db  'Uninstalled',0dh,0ah,'$'

cseg ends

end main
